/*
 * README
 *
 * This is an example to show you how the APIs of SEfile can be used.
 * First we create an encrypted text file, then we retrieve its content and print it.
 * Then we also show you how to use SEfile with encrypted SQLite databases.
 *
*/

#include "../sefile/SEfile.h"
#include "../sqlite/sqlite3.h"
#include "../sefile/environment.h"

using namespace std;

int main(){
	unique_ptr<L0> l0 = make_unique<L0>();
	unique_ptr<L1> l1 = make_unique<L1>();

	uint8_t numDevices = l0->GetNumberDevices(); // check how many SEcube(s) are connected
	if(numDevices != 1){
		cout << "SEcube not connected or too many SEcube connected. Exit." << endl;
		return -1;
	}

	uint8_t pin[32]; // we need a PIN to login to the SEcube, there are 2 PINs: user and admin
	memset(pin, 0, 32); // imagine that the PIN we need to use is made of all zeros
	try{
		l1->L1Login(pin, SE3_ACCESS_ADMIN, true); // here we access as administrator but you can normally access as user
		cout << "SEcube logged in." << endl;
	} catch(...){
		cout << "Wrong PIN. Exit." << endl;
		return -1;
	}

	/* Now we want to create a simple text file encrypted with SEfile. Therefore, instead of using the normal functions provided by the
	 * OS, we will use the functions of SEfile. */
	uint32_t keyID = 1; // this is the key that we want to use to encrypt the file
	SEfile file1(l1.get(), keyID, L1Algorithms::Algorithms::AES_HMACSHA256); // we create a SEfile object bounded to the L1 SEcube object, we specify also the ID of the key and the algorithm that we want to use
	string filename = "sefile_example.txt";
	string content = "Lorem ipsum dolor sit amet, consectetur adipiscing elit. Nullam arcu ex, interdum in dolor quis, semper laoreet dolor. "
					 "Etiam eros mauris, interdum sed consectetur sit amet, commodo eget massa. Pellentesque tincidunt ullamcorper rhoncus. "
					 "Proin eget congue velit, eu faucibus libero. Mauris cursus in ante sit amet blandit. Fusce sit amet.";
	/* here we simply call the SEfile open function, notice that we pass the SEFILE_WRITE flag because we want to modify the content of the file, then we also
	 * pass the SEFILE_NEWFILE flag because we want to create the file */
	file1.secure_open((char*)filename.c_str(), SEFILE_WRITE, SEFILE_NEWFILE);
	int pos; /* we move to the end of the file (this is not really necessary here because the file has just been created) */
	file1.secure_seek(0, &pos, SEFILE_END); // append to the end of the file
	file1.secure_write((uint8_t*)content.c_str(), content.size()); /* now we simply write our buffer to the file */
	file1.secure_close(); // finally we close the file

	/* Now suppose that we want to read the content of the file we have just created */
	SEfile file2(l1.get()); // we create another SEfile object (we could have used the previous one)
	/* notice that we set only the L1 object for file2, we do not set the key or the algorithm. we could have set also those parameters because we know them but, tipically,
	 * in the real world you don't know which key and which algorithm have been used to encrypt a file protected by SEfile. this is due to the fact that you cannot keep track
	 * of the encryption details of many, many files. for this reason SEfile will automatically understand, from the header of the encrypted file, which key ID and which algorithm
	 * have been used...so there is no need to set them in advance when you simply want to open an existing file (you must do that only when you want to create a file) */
	file2.secure_open((char*)filename.c_str(), SEFILE_READ, SEFILE_OPEN); // notice that we simply read the file and we open it because it already exists
	file2.secure_seek(0, &pos, SEFILE_BEGIN); // not really necessary, when a file is opened its pointer is already set to the first sector (header excluded)
	unique_ptr<char[]> filecontent;
	uint32_t filedim;
	secure_getfilesize((char*)filename.c_str(), &filedim, l1.get()); // here we retrieve the total size of the valid data stored inside the file
	filecontent = make_unique<char[]>(filedim);
	unsigned int bytesread;
	file2.secure_read((uint8_t*)filecontent.get(), filedim, &bytesread); // here we simply read the entire file into our buffer
	file2.secure_close(); // we can close our file
	string readcontent(filecontent.get(), filedim);
	cout << readcontent << endl; // we print the content of the text file

	/* Now we show you how to use SEfile to create an encrypted SQLite database */
	sqlite3 *db;
	keyID = 1; // this is the key that we want to use to encrypt the file
	unique_ptr<SEfile> mydatabase = make_unique<SEfile>(); // notice that we use a standard SEfile object also for encrypted databases
	mydatabase->secure_init(l1.get(), keyID, L1Algorithms::Algorithms::AES_HMACSHA256); // we initialize the SEfile object with the parameters we want (we could have done this using the specific constructor)
	char dbname[] = "my_encrypted_database";
	memcpy(mydatabase->handleptr->name, dbname, strlen(dbname)); // you MUST copy the name of the database that you want to open/create into the name field of the handleptr attribute of the SEfile object
	databases.push_back(std::move(mydatabase)); // you MUST move the SEfile object to the databases vector
	sqlite3_open(dbname, &db); // now you can open your database
	/* here you can do whatever you want with the database, always using the standard APIs of the SQLite C interface */
	sqlite3_close(db); // finally you close your database

	l1->L1Logout();
	return 0;
}
